import sys

from carnival.step import Step
from carnival.host import SSHHost, LocalHost
from carnival.task import TaskBase, StepsTask
from carnival import cmd
from carnival import internal_tasks
from carnival.utils import log
from carnival.context import context_ref


if not sys.warnoptions:
    import warnings
    warnings.filterwarnings(
        action='default',
        module=r'carnival.*'
    )


__all__ = [
    'Step',
    'SSHHost', 'LocalHost',
    'TaskBase', 'StepsTask',
    'cmd',
    'log',
    'context_ref',
    'internal_tasks',
]



"""
Объект, представляющий единицу оборудования.

Carnival не предоставляет никаких сложных абстракций для работы с группами хостов,
подразумевая что вы будете использовать встроенные коллекции python и организуете
работу так, как будет удобно для вашей задачи.
В простом случае, можно передавать хосты прямо в коде файла `carnival_tasks.py`.
В более сложных, создать списки в отдельном файле, например `inventory.py`
"""

import typing

from fabric.connection import Connection as SSHConnection  # type: ignore
from invoke.context import Context as LocalConnection  # type: ignore
from paramiko.client import MissingHostKeyPolicy, AutoAddPolicy


AnyConnection = typing.Union[SSHConnection, LocalConnection]


class LocalHost:
    """
    Локальный хост, работает по локальному терминалу

    :param context: Контекст хоста
    """

    def __init__(self, **context: typing.Any) -> None:
        self.addr = "local"
        self.context = context
        self.context['host'] = self

    def connect(self) -> LocalConnection:
        return LocalConnection()

    def __str__(self) -> str:
        return f"🖥 {self.addr}"

    def __hash__(self) -> int:
        return hash(self.addr)

    def __repr__(self) -> str:
        return f"<Host object {self.addr}>"


class SSHHost(LocalHost):
    """
    Удаленный хост, работает по SSH
    """

    def __init__(
        self,
        addr: str,
        ssh_user: typing.Optional[str] = None, ssh_password: typing.Optional[str] = None, ssh_port: int = 22,
        ssh_gateway: typing.Optional['SSHHost'] = None,
        ssh_connect_timeout: int = 10,
        missing_host_key_policy: typing.Type[MissingHostKeyPolicy] = AutoAddPolicy,

        **context: typing.Any
     ):
        """
        :param addr: Адрес сервера
        :param ssh_user: Пользователь SSH
        :param ssh_password: Пароль SSH
        :param ssh_port: SSH порт
        :param ssh_connect_timeout: SSH таймаут соединения
        :param ssh_gateway: Gateway
        :param context: Контекст хоста
        """
        if ":" in addr:
            raise ValueError("Please set port in 'ssh_port' arg")

        if "@" in addr:
            raise ValueError("Please set user in 'ssh_user' arg")

        super().__init__(**context)

        self.addr = addr
        self.ssh_port = ssh_port
        self.ssh_user = ssh_user
        self.ssh_password = ssh_password
        self.ssh_connect_timeout = ssh_connect_timeout
        self.ssh_gateway: typing.Optional['SSHHost'] = ssh_gateway
        self.missing_host_key_policy = missing_host_key_policy

    def connect(self) -> SSHConnection:
        gateway = None
        if self.ssh_gateway:
            gateway = self.ssh_gateway.connect()
            assert isinstance(gateway, SSHConnection), f"{self.ssh_gateway} is not ssh connection"

        conn = SSHConnection(
            host=self.addr,
            port=self.ssh_port,
            user=self.ssh_user,
            connect_timeout=self.ssh_connect_timeout,
            gateway=gateway,
            connect_kwargs={
                'password': self.ssh_password,
            }
        )
        conn.client.set_missing_host_key_policy(self.missing_host_key_policy)
        return conn


AnyHost = typing.Union[LocalHost, SSHHost]



from carnival import TaskBase


class Help(TaskBase):
    """
    Показать список доступных задач
    """
    module_name = ""
    help = "List available tasks and help"

    def run(self) -> None:
        from carnival.cli import task_types

        task_list = list(task_types.keys())
        task_list.sort()

        ralign = 4
        if task_list:
            ralign += max([len(x) for x in task_list])

        for task_name in task_list:
            help_text = task_types[task_name].help
            if help_text:
                print(f" {task_name:<{ralign}} {help_text}")
            else:
                print(f" {task_name}")


class Validate(TaskBase):
    """
    Запустить валидацию доступых задач и напечатать список ошибок
    """
    module_name = ""
    help = "Validate available tasks"

    def run(self) -> None:
        from carnival.cli import task_types

        all_errors = []

        task_list = list(task_types.keys())
        task_list.sort()

        for task_name in task_list:
            all_errors += task_types[task_name]().validate()

        if all_errors:
            for e in all_errors:
                print(f" * {e}")
            return

        print("All tasks are valid.")



import abc
import re
import copy
import typing

from carnival import Step, connection
from carnival.host import AnyHost
from carnival.exceptions import ContextBuilderError


def _underscore(word: str) -> str:
    # https://github.com/jpvanhal/inflection/blob/master/inflection.py
    word = re.sub(r"([A-Z]+)([A-Z][a-z])", r'\1_\2', word)
    word = re.sub(r"([a-z\d])([A-Z])", r'\1_\2', word)
    word = word.replace("-", "_")
    return word.lower()


class TaskBase:
    """
    Задача это единица выполнения одного или несколькоих шагов на определенных хостах.

    Именование задач.

    Полное имя задачи состоит из двух частей. <module_name>.<name>.
    carnival автоматически генерирует имена задач из этих частей, но есть возможность управлять этим вручную,
    используя два атрибута класса Task.

    name:
    module_name:

    >>> class CheckDiskSpace(TaskBase):
    >>> help = "Print server root disk usage"
    >>>
    >>> def run(self, disk: str = "/") -> None:
    >>>    with connection.SetConnection(my_server):
    >>>        cmd.cli.run(f"df -h {disk}", hide=False)

    """

    # Имя задачи
    name: str = ""
    """
    название задачи. если не определено имя будет сгенерировано автоматически.
    """
    module_name: typing.Optional[str] = None
    """
    имя модуля. если назначить пустую строку, полное имя будет включать только название задачи.
    """

    help: str = ""
    """
    Строка помощи при вызове carnival help
    """

    @classmethod
    def get_name(cls) -> str:
        return cls.name if cls.name else _underscore(cls.__name__)

    def call_task(self, task_class: typing.Type['TaskBase']) -> typing.Any:
        """
        Запустить другую задачу
        Возвращает результат работы задачи
        """
        return task_class().run()

    def validate(self) -> typing.List[str]:
        """
        Хук для проверки валидности задачи перед запуском, не вызывается автоматически
        """

        return []

    @abc.abstractmethod
    def run(self) -> None:
        """
        Реализация выполнения задачи
        """
        raise NotImplementedError


class StepsTask(abc.ABC, TaskBase):
    """
    Запустить шаги `steps` на хостах `hosts`

    >>> class InstallPackages(StepsTask):
    >>>    help = "Install packages"
    >>>
    >>>    hosts = [my_server]
    >>>    steps = [InstallStep()]

    """

    hosts: typing.List[AnyHost]
    """
    Список хостов
    """
    steps: typing.List[Step]
    """
    Список шагов в порядке выполнения
    """

    def extend_host_context(self, host: AnyHost) -> typing.Dict[str, typing.Any]:
        """
        Метод для переопределения контекста хоста

        :param host: хост на котором готовится запуск
        """
        return copy.deepcopy(host.context)

    def validate(self) -> typing.List[str]:
        """
        Хук для проверки валидности задачи перед запуском, проверяет примеримость контекста хостов на шагах
        """

        errors: typing.List[str] = []

        for host in self.hosts:
            for step in self.steps:
                try:
                    step.run_with_context(self.extend_host_context(host=host))
                except ContextBuilderError as ex:
                    errors.append(f"{self.__class__.__qualname__} -> {step.__class__.__qualname__} on {host}: {ex}")

        return errors

    def run(self) -> None:
        errors = self.validate()

        if errors:
            print("There is context building errors")
            for e in errors:
                print(f" * {e}")
            return

        for host in self.hosts:
            host_ctx = self.extend_host_context(host=host)
            with connection.SetConnection(host):
                for step in self.steps:
                    step_name = _underscore(step.__class__.__name__)
                    print(f"💃💃💃 Running {self.get_name()}:{step_name} at {host}")
                    call_step = step.run_with_context(host_ctx)
                    call_step()



import inspect
import os
import typing

from carnival.exceptions import ContextBuilderError, ContextBuilderPassAllArgs


def _get_arg_names(fn: typing.Callable[..., typing.Any]) -> typing.Dict[str, bool]:
    arg_names: typing.Dict[str, bool] = {}

    for arg_name, arg_parameter in inspect.signature(fn).parameters.items():
        if arg_parameter.kind not in [
            arg_parameter.KEYWORD_ONLY,
            arg_parameter.POSITIONAL_OR_KEYWORD,
            arg_parameter.VAR_KEYWORD,
        ]:
            raise ContextBuilderError("only keyword parameters required for autocontext")

        if arg_parameter.kind == arg_parameter.VAR_KEYWORD:
            raise ContextBuilderPassAllArgs()

        arg_names[arg_name] = arg_parameter.default == arg_parameter.empty

    if 'self' in arg_names:
        arg_names.pop('self')

    return arg_names


def build_kwargs(
    fn: typing.Callable[..., typing.Any],
    context: typing.Dict[str, typing.Any],
) -> typing.Dict[str, typing.Any]:
    try:
        arg_names: typing.Dict[str, bool] = _get_arg_names(fn)
    except ContextBuilderPassAllArgs:
        # Pass all context if kwargs var exists
        return context.copy()

    kwargs = {}
    for arg_name, is_arg_required in arg_names.items():
        if arg_name not in context and is_arg_required:
            raise ContextBuilderError(f"Variable '{arg_name}', is not present in context.")

        if arg_name in context:
            kwargs[arg_name] = context[arg_name]
    return kwargs


def build_context(*context_chain: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
    run_context: typing.Dict[str, typing.Any] = {}

    env_prefix = "CARNIVAL_CTX_"
    # Build context from environment variables
    for env_name, env_val in os.environ.items():
        if env_name.startswith(env_prefix):
            run_context[env_name[len(env_prefix):]] = env_val

    for context_item in context_chain:
        for var_name, var_val in context_item.items():
            if isinstance(var_val, context_ref):
                try:
                    run_context[var_name] = run_context[var_val.context_var_name]
                except KeyError as e:
                    raise ContextBuilderError(f"There is no '{var_val.context_var_name}' variable in context") from e
            else:
                run_context[var_name] = var_val

    return run_context


class context_ref:
    """
    Ссылка на другую переменную контекста

    Например, так можно передать в переменную c именем `name` `Step` другую переменную контекста с именем `dns_domain`
    >>> TestStep(
    >>>   name=context_ref('dns_domain'),
    >>> ),

    """
    def __init__(self, context_var_name: str):
        self.context_var_name = context_var_name





class CarnivalException(BaseException):
    """
    Base error class
    """


class ContextBuilderError(CarnivalException):
    """
    Error when build context for step
    """


class ContextBuilderPassAllArgs(CarnivalException):
    """
    Signal to send full context to step with **kwargs
    """


class GlobalConnectionError(CarnivalException):
    """
    Global connection switching error
    """
